/*
 * File event notification support.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <map>
#include <mutex>

#include <ipc/cap_alloc.h>
#include <ipc/server.h>

#include <pthread.h>
#include <pthread-l4.h>

#include "notification_client.h"
#include "notifier.h"



/* Thread-local storage workaround. */

static std::mutex lock;
static std::map<l4_cap_idx_t, FileNotifier *> notifiers;

FileNotifier *get_notifier()
{
  std::lock_guard<std::mutex> guard(lock);

  l4_cap_idx_t thread = pthread_l4_cap(pthread_self());
  FileNotifier *notifier = notifiers[thread];

  /* Start any new notifier. */

  if (notifier == NULL)
  {
    notifier = new FileNotifier;
    notifiers[thread] = notifier;
    notifier->start();
  }

  return notifier;
}



/* Subscribe to notification events on a file. */

long FileNotifier::subscribe(file_t *file, notify_flags_t flags)
{
  /* Create a notification endpoint, if necessary. */

  if (l4_is_invalid_cap(file->notifier))
  {
    long err = ipc_server_new_for_thread(&file->notifier, file, _thread);

    if (err)
      return err;
  }

  client_Notification notify(file->ref);

  return notify.subscribe(file->notifier, flags);
}

/* Unsubscribe from notification events on a file. */

long FileNotifier::unsubscribe(file_t *file)
{
  if (l4_is_invalid_cap(file->notifier))
    return -L4_EINVAL;

  ipc_cap_free_um(file->notifier);

  client_Notification notify(file->ref);

  return notify.unsubscribe();
}

/* Handle a notification event for a file. Ideally, this would be invoked by the
   generic server dispatch mechanism, with the gate label being interpreted and
   provided as the first parameter. */

void FileNotifier::_notify(file_t *file, notify_flags_t flags)
{
  std::unique_lock<std::mutex> guard(_lock);

  /* Record the flags in the file object. */

  file->notifications = flags;
  _affected.push_back(file);

  /* Notify any waiting caller. */

  _notified.notify_one();
}

/* Listen for notifications. */

void FileNotifier::mainloop()
{
  ipc_message_t msg;
  l4_umword_t label;

  while (1)
  {
    ipc_message_wait(&msg, &label);

    /* Clear lower label bits. */

    label = label & ~3UL;

    /* Ignore erroneous messages. */

    if (l4_ipc_error(msg.tag, l4_utcb()))
      continue;

    /* Reply to notifications. */

    ipc_message_reply(&msg);

    /* Interpret gate labels as file objects. */

    file_t *file = (file_t *) label;
    notify_flags_t flags = ipc_message_get_word(&msg, 0);

    /* Register the notification. */

    _notify(file, flags);
  }

  ipc_message_free(&msg);
}

/* Invoke the mainloop in a thread. */

static void *notifier_mainloop(void *data)
{
  FileNotifier *notifier = reinterpret_cast<FileNotifier *>(data);

  notifier->mainloop();
  return 0;
}

/* Start listening for notifications. */

long FileNotifier::start()
{
  if (_started)
    return L4_EOK;

  pthread_t thread;
  pthread_attr_t attr;
  long err;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  err = pthread_create(&thread, &attr, notifier_mainloop, this);
  if (err)
    return err;

  _thread = pthread_l4_cap(thread);
  _started = true;

  return L4_EOK;
}

/* Wait for notification events on files. */

long FileNotifier::wait(file_t **file)
{
  std::unique_lock<std::mutex> guard(_lock);

  while (1)
  {
    if (!_affected.empty())
    {
      *file = _affected.front();
      _affected.pop_front();
      return L4_EOK;
    }
    else
      _notified.wait(guard);
  }

  return L4_EOK;
}

// vim: tabstop=2 expandtab shiftwidth=2
